-- ------------------------
-- KAPITEL 10 (PostGIS 3.1)
-- ------------------------

-- 10.4 Topologie in PostGIS

-- Falls ein Benutzerkonto keine Rechte als Datenbankadministrator besitzt, muss dieser die notwendigen Zugriffsrechte gewähren:

GRANT USAGE ON SCHEMA topology TO <DB-Benutzerkonto>;
GRANT ALL ON ALL TABLES IN SCHEMA topology TO <DB-Benutzerkonto>;
GRANT USAGE, SELECT ON ALL SEQUENCES IN SCHEMA topology TO <DB-Benutzerkonto>;

-- 10.4.1 Topologieschema

SELECT topology.CreateTopology('geodbland_topo', 0, 0.001);
SELECT * FROM topology.Topology;

SELECT * FROM geodbland_topo.face;

-- Schritt 1: Masche f1 erzeugen: 
SELECT topology.ST_AddIsoNode('geodbland_topo',0,ST_Point(1,1));
SELECT topology.ST_AddEdgeModFace('geodbland_topo',1,1,
   ST_LineFromText('LINESTRING(1 1, 19 1, 19 20, 1 20, 1 1)'));

-- Schritt 2: Masche f2 erzeugen: 
SELECT topology.ST_AddIsoNode('geodbland_topo',1,ST_Point(3,10));
SELECT topology.ST_AddEdgeModFace('geodbland_topo',2,2,ST_LineFromText(
   'LINESTRING(3 10,4 5,9 3,13 5,16 8,17 12,13 15,7 14,3 10)')); 

-- Schritt 3: Kante e5 einziehen: 
SELECT topology.ST_ModEdgeSplit('geodbland_topo',1,ST_Point(19,12));
SELECT topology.ST_ModEdgeSplit('geodbland_topo',2,ST_Point(17,12));
SELECT topology.ST_AddEdgeModFace('geodbland_topo',3,4,
   ST_LineFromText('LINESTRING(19 12, 17 12)'));

-- Schritt 4: Kante e6 einführen und Masche f3 erzeugen: 
SELECT topology.ST_ModEdgeSplit('geodbland_topo',3,ST_Point(1,10));
SELECT topology.ST_AddEdgeModFace('geodbland_topo',2,5,
   ST_LineFromText('LINESTRING(3 10, 1 10)')); 

-- Validierung: 
SELECT topology.ValidateTopology('geodbland_topo'); 

-- Abfrage der topolgischen Primitive aus dem Schema geodbland_topo: 
SELECT * FROM geodbland_topo.node;
SELECT * FROM geodbland_topo.edge;
SELECT * FROM geodbland_topo.face;

-- Bestimmung der Maschengeometrien (ohne Masche f0): 
SELECT face_id, topology.ST_GetFaceGeometry('geodbland_topo',face_id)
FROM geodbland_topo.face WHERE face_id > 0;

SELECT topology.TopoGeo_AddPolygon('geodbland_topo',
                                   ST_GeometryN(geo,2), 0.001)
FROM GeoDbLand WHERE name = 'Land';

SELECT topology.TopoGeo_AddLineString('geodbland_topo',
                                      ST_Boundary(geo), 0.001)
FROM GeoDbLand WHERE name = 'Haus';

SELECT topology.GetNodeByPoint('geodbland_topo',ST_Point(7,7),0.1);
SELECT topology.GetEdgeByPoint('geodbland_topo',ST_Point(7,7),0.1);

SELECT topology.GetFaceByPoint('geodbland_topo',ST_Point(7,7),0.1);

-- 10.4.2 Datentyp TopoGeometry

CREATE TABLE GeoDbLandNutzung (
  id       INTEGER,                          -- ID
  nutzung  VARCHAR(40),                      -- Landnutzungsart
  CONSTRAINT pk_gdblandnutz PRIMARY KEY(id)  -- Primärschlüssel
);

SELECT topology.AddTopoGeometryColumn('geodbland_topo', 'public',
       'geodblandnutzung', 'gebiet', 'POLYGON');

SELECT topology_id, layer_id, schema_name,
       table_name, feature_column, feature_type FROM topology.Layer;

-- Vier Landnutzungen aus den fünf vorhandenen Maschen erzeugen: 
INSERT INTO GeoDbLandNutzung VALUES (1, 'Wald',
  topology.CreateTopoGeom('geodbland_topo', 3, 1, ARRAY[ARRAY[3,3]]));
INSERT INTO GeoDbLandNutzung VALUES (2, 'Wasser',
  topology.CreateTopoGeom('geodbland_topo', 3, 1, ARRAY[[2,3]]));
INSERT INTO GeoDbLandNutzung VALUES (3, 'Bebauung',
  topology.CreateTopoGeom('geodbland_topo', 3, 1, ARRAY[[5,3]]));
INSERT INTO GeoDbLandNutzung VALUES (4, 'Weideland', topology.CreateTopoGeom(
  'geodbland_topo', 3, 1, ARRAY[[1,3],[4,3]]));

-- Eine Landnutzung einfügen, die eine neue Masche erzeugt: 
INSERT INTO GeoDbLandNutzung
VALUES (5, 'Sumpf', topology.toTopoGeom(
  ST_PolygonFromText('POLYGON((1 20,0 20,0 0,19 0,19 1,1 1,1 20))'),
  'geodbland_topo', 1, 0.001));

SELECT nutzung, CAST(gebiet AS GEOMETRY) FROM GeoDbLandNutzung;

SELECT topology.AsGML(gebiet) FROM GeoDbLandNutzung
WHERE nutzung = 'Bebauung';

-- Die Maschen-IDs zu GeoDbLandNutzung: 
SELECT element_id
FROM geodbland_topo.relation r INNER JOIN GeoDbLandNutzung
ON (gebiet).id = r.topogeo_id AND (gebiet).layer_id = r.layer_id;

-- Die Maschen-Geometrien zu GeoDbLandNutzung: 
SELECT element_id, topology.ST_GetFaceGeometry('geodbland_topo',element_id)
FROM geodbland_topo.relation r INNER JOIN GeoDbLandNutzung
ON (gebiet).id = r.topogeo_id AND (gebiet).layer_id = r.layer_id;

SELECT l1.nutzung AS nutzung1, l2.nutzung AS nutzung2,
       topology.Intersects(l1.gebiet,l2.gebiet)
FROM GeoDbLandNutzung l1 INNER JOIN GeoDbLandNutzung l2
ON l1.nutzung < l2.nutzung
ORDER BY nutzung1, nutzung2;

SELECT nutzung, ST_Simplify(gebiet,1.4) FROM GeoDbLandNutzung;

-- Die Insel als Weideland entfernen: 
UPDATE GeoDbLandNutzung
SET gebiet = topology.TopoGeom_remElement(gebiet,ARRAY[4,3])
WHERE nutzung = 'Weideland';

--- Die Insel der Waldnutzung zuschlagen: 
UPDATE GeoDbLandNutzung
SET gebiet = topology.TopoGeom_addElement(gebiet,ARRAY[4,3])
WHERE nutzung = 'Wald';

CREATE INDEX GeoDbLandNutzung_geom_ix ON GeoDbLandNutzung
USING GIST (gebiet);

CREATE INDEX GeoDbLandNutzung_geom_ix ON GeoDbLandNutzung
USING GIST (CAST(gebiet AS Geometry));

-- Materialisierte Sicht mit Geometrieattribut einführen: 
CREATE MATERIALIZED VIEW GeoDbLandNutzungView AS
  SELECT *, CAST(gebiet AS Geometry) AS geom FROM GeoDbLandNutzung;

-- Index anlegen: 
CREATE INDEX GeoDbLandNutzungView_geom_ix
ON GeoDbLandNutzungView USING GIST (geom);
-- Punktanfrage mit Flaggenmast: 
SELECT nutzung, GeoDbLandNutzungView.geom
FROM GeoDbLandNutzungView, GeoDbLand
WHERE ST_Contains(GeoDbLandNutzungView.geom, GeoDbLand.geo) AND
      GeoDbLand.name = 'Flaggenmast';

REFRESH MATERIALIZED VIEW GeoDbLandNutzungView;

SELECT nutzung, CAST(gebiet AS GEOMETRY)
FROM GeoDbLandNutzung, geodbland_topo.relation AS r, GeoDbLand
WHERE r.topogeo_id = (gebiet).id AND r.layer_id = (gebiet).layer_id AND
   GeoDbLand.name = 'Flaggenmast' AND 
   r.element_id = topology.GetFaceByPoint('geodbland_topo',GeoDbLand.geo,0);
