-- ------------------------
-- KAPITEL 11 (PostGIS 3.1)
-- ------------------------

-- 11.3.1 Erzeugung und Speicherung von Rasterdaten

CREATE TABLE GeoRasterkarten (
  id         INTEGER   GENERATED ALWAYS AS IDENTITY,
  quelle     VARCHAR(255),                -- Name der Quelldatei
  georaster  RASTER,                      -- das Raster
  CONSTRAINT pk_georst PRIMARY KEY (id)   -- Primärschlüssel
);

-- 11.3.1.1 Rasterdaten über raster2pgsql

-- Hier über raster2pgsql sentinel2.tif speichern.

SELECT * FROM GeoRasterkarten;

-- Hier über raster2pgsql vg250_gem.png speichern.

SELECT id, quelle, ST_GeoReference(georaster) FROM GeoRasterkarten;

-- 11.3.1.2 Rasterdaten aus BLOBs

-- Hier (spätestens) Karte wms_dtk500.png über Java als BLOB speichern.

INSERT INTO GeoRasterkarten (georaster,quelle)
  SELECT ST_FromGDALRaster(karte,31467), name
  FROM Rasterkarten WHERE id = 1;

UPDATE GeoRasterkarten
SET georaster = ST_SetGeoReference(georaster,
                (SELECT worldfile FROM Rasterkarten WHERE id = 1), 'ESRI')
WHERE quelle = 'wms_dtk500.png';

-- 11.3.1.3 Rasterdaten aus GEOMETRY

SELECT geo
FROM (SELECT ST_Transform(Geometry(geog),31467) AS geo FROM Ortsnetze) sq
WHERE geo && ST_MakeEnvelope(3430000,5870000, 3470000,5910000);

INSERT INTO GeoRasterkarten (quelle,georaster)
  SELECT 'ortsnetz-area', ST_AsRaster(ST_Collect(geo), 75.0,-75.0,
    ARRAY['8BUI','8BUI','8BUI'], ARRAY[100,200,100], ARRAY[0,0,0])
  FROM (SELECT ST_Transform(Geometry(geog),31467) AS geo FROM Ortsnetze) sq
  WHERE geo && ST_MakeEnvelope(3430000,5870000, 3470000,5910000);

INSERT INTO GeoRasterkarten (quelle,georaster)
  SELECT 'ortsnetz-border', ST_AsRaster(ST_Collect(ST_Boundary(geo)),
    75.0,-75.0, ARRAY['8BUI','8BUI','8BUI'], ARRAY[200,20,20], ARRAY[0,0,0])
  FROM (SELECT ST_Transform(Geometry(geog),31467) AS geo FROM Ortsnetze) sq
  WHERE geo && ST_MakeEnvelope(3430000,5870000, 3470000,5910000);

-- 11.3.1.4 Metadaten und Beschränkungen

SELECT r_table_name, r_raster_column, srid FROM Raster_Columns;

SELECT AddRasterConstraints('georasterkarten','georaster','srid');

SELECT r_table_name, r_raster_column, srid FROM Raster_Columns;

-- 11.3.2.1 Abfrage von Eigenschaften

SELECT quelle, ST_Width(georaster), ST_Height(georaster),
       ST_SRID(georaster), ST_NumBands(georaster), ST_MemSize(georaster)
FROM GeoRasterkarten;

SELECT quelle, ST_UpperLeftX(georaster), ST_UpperLeftY(georaster),
               ST_ScaleX(georaster), ST_ScaleY(georaster), 
               ST_SkewX(georaster), ST_SkewY(georaster)
FROM GeoRasterkarten;

SELECT quelle, ST_Count(georaster,1) FROM GeoRasterkarten;

SELECT (ST_ValueCount(georaster,1)).*
FROM GeoRasterkarten
WHERE quelle = 'wms_dtk500.png'
ORDER BY (ST_ValueCount(georaster,1)).count DESC;

-- 11.3.2.2 Umrechnung zwischen Welt- und Pixelkoordinaten

SELECT (ST_RasterToWorldCoord(georaster,16,28)).*,
        ST_RasterToWorldCoordX(georaster,16),
        ST_RasterToWorldCoordY(georaster,28)
FROM GeoRasterkarten WHERE quelle = 'wms_dtk500.png';

SELECT (ST_WorldToRasterCoord(georaster,3431000,5908200)).*,
        ST_WorldToRasterCoordX(georaster,3431000),
        ST_WorldToRasterCoordY(georaster,5908200)
FROM GeoRasterkarten WHERE quelle = 'wms_dtk500.png';

-- 11.3.2.3 Zugriff auf Rasterdaten

SELECT ST_Value(georaster,1,ST_SetSRID(ST_MakePoint(3431000,5908200),31467)),
       ST_Value(georaster,1,16,28)
FROM GeoRasterkarten WHERE quelle = 'wms_dtk500.png';

SELECT ST_Value(georaster,1,ST_SetSRID(ST_MakePoint(3431000,5908200),31467)),
       ST_Value(georaster,1,ST_SetSRID(ST_MakePoint(3431000,5908200),31467),
                false)
FROM GeoRasterkarten WHERE quelle = 'ortsnetz-border';

SELECT ST_NearestValue(georaster,1,
                       ST_SetSRID(ST_MakePoint(3431000,5908200),31467))
FROM GeoRasterkarten WHERE quelle = 'ortsnetz-border';

SELECT (ST_PixelOfValue(georaster,1,100)).*
FROM GeoRasterkarten WHERE quelle = 'ortsnetz-area';

SELECT ST_PixelAsPolygon(georaster,16,28)
FROM GeoRasterkarten WHERE quelle = 'wms_dtk500.png';

SELECT (ST_PixelAsCentroids(georaster,1)).*
FROM GeoRasterkarten WHERE quelle = 'ortsnetz-border';

SELECT (ST_DumpAsPolygons(georaster,1)).*
FROM GeoRasterkarten WHERE quelle = 'ortsnetz-area';

-- 11.3.2.4 Kachelung und räumliche Indexierung

CREATE TABLE DTK500 (
  id         INTEGER   GENERATED ALWAYS AS IDENTITY,
  kachel     RASTER,                      -- die Rasterkachel
  CONSTRAINT pk_dtk500 PRIMARY KEY (id)   -- Primärschlüssel
);

INSERT INTO DTK500 (kachel)
  SELECT ST_Tile(georaster,75,75,true) 
  FROM georasterkarten WHERE quelle = 'wms_dtk500.png';

CREATE INDEX dtk_kachel_ix ON DTK500 USING GIST(ST_Envelope(kachel));

SELECT ST_Transform(geom,31467), ST_Envelope(kachel)
FROM DTK500 INNER JOIN GeoGemeinden
ON ST_Contains(ST_Envelope(kachel),ST_Transform(geom,31467))
WHERE code = 'DE03403000';

SELECT count(id), ST_Union(kachel)
FROM DTK500 INNER JOIN Ortsnetze
ON ST_Intersects(ST_Envelope(kachel),ST_Transform(Geometry(geog),31467))
WHERE vorwahl = '0441';

-- 11.3.2.5 Rasterverarbeitung und Bildpyramide

ALTER TABLE DTK500 ADD level INTEGER;

UPDATE DTK500 SET level = 1;

INSERT INTO DTK500 (kachel, level)
  SELECT ST_Resize(ST_Tile(georaster,150,150,true),75,75,'CUBIC'),2
  FROM georasterkarten WHERE quelle = 'wms_dtk500.png';
INSERT INTO DTK500 (kachel, level)
  SELECT ST_Resize(ST_Tile(georaster,300,300,true),75,75,'CUBIC'),3
  FROM georasterkarten WHERE quelle = 'wms_dtk500.png';
INSERT INTO DTK500 (kachel, level)
  SELECT ST_Resize(georaster,75,75,'CUBIC'),4
  FROM georasterkarten WHERE quelle = 'wms_dtk500.png';

SELECT ST_Union(georaster,'LAST') FROM
  (SELECT georaster FROM georasterkarten
   WHERE quelle LIKE 'ortsnetz%' ORDER BY quelle) sq;

-- Resampling: 
INSERT INTO GeoRasterkarten (quelle,georaster)
VALUES ('ortsnetz-border resamp', ST_Resample(
 (SELECT georaster FROM GeoRasterkarten WHERE quelle='ortsnetz-border'),
 (SELECT georaster FROM GeoRasterkarten WHERE quelle='sentinel2.tif'),
 'CUBIC') );

-- Prüfung der Pixelgrößen: 
SELECT quelle, ST_PixelWidth(georaster), ST_PixelWidth(georaster)
FROM GeoRasterkarten
WHERE quelle LIKE 'ortsnetz-border%';

-- Anpassung des Ausschnitts: 
UPDATE GeoRasterkarten SET georaster = ST_Clip(georaster,
  (SELECT ST_Envelope(georaster) FROM georasterkarten 
   WHERE quelle='sentinel2.tif'), true)
WHERE quelle = 'ortsnetz-border resamp';

-- Zusammenführung: 
SELECT ST_Union(georaster,'FIRST') FROM
  (SELECT georaster FROM georasterkarten
   WHERE quelle = 'ortsnetz-border resamp' OR 
         quelle = 'sentinel2.tif' ORDER BY quelle) sq;

SELECT ST_Clip(ST_Union(kachel),ST_Transform(Geometry(geog),31467),true)
FROM DTK500 INNER JOIN Ortsnetze
ON ST_Intersects(ST_Envelope(kachel),ST_Transform(Geometry(geog),31467))
WHERE vorwahl = '0441' AND level = 1
GROUP BY geog;

		 