-- ------------------------
-- KAPITEL 12 (PostGIS 3.1)
-- ------------------------

-- 12.3.1.1 Punkte, Linienzüge und Polygone in 3D

-- Anlegen der Tabelle:
CREATE TABLE GeoDbLand3D (
  id     INTEGER,                          -- Identifikationsnummer
  name   VARCHAR(20),                      -- Objektname
  geo    GEOMETRY(GEOMETRYZ,0),            -- 3D-Geometrie
  CONSTRAINT pk_gdbland3D PRIMARY KEY(id)  -- Primärschlüssel
);

-- Einfügen der Daten:
INSERT INTO GeoDbLand3D (id,name,geo)
VALUES (1, 'Flaggenmast', ST_PointFromText('POINT Z (10 11 5)'));

INSERT INTO GeoDbLand3D (id,name,geo)
VALUES (2, 'Steg', ST_LineFromText('LINESTRING Z (4 4 1,5 7 2.5,8 8 2)', 0));

INSERT INTO GeoDbLand3D (id,name,geo)
VALUES (3, 'Baumreihe',
  ST_MPointFromText('MULTIPOINT Z (10 19 5,12 19 10,14 19 7)'));
  
INSERT INTO GeoDbLand3D (id,name,geo)
VALUES (6, 'Land', 
  ST_MPolyFromText('MULTIPOLYGON Z ( ( (1 1 1,19 1 1,19 20 1,1 20 1,1 1 1),
      (4 5 1,3 10 1,7 14 1,13 15 1,17 12 1,16 8 1,13 5 1,9 3 1,4 5 1) ),
    ( (7 7 2,12 7 2,13 11 2,10 13 2,7 10 2,7 7 2) ) )'));

-- Koordinatenwerte und -dimension:
SELECT ST_Z(geo), ST_ZMin(geo), ST_CoordDim(geo), ST_ZMflag(geo)
FROM GeoDbLand3D
WHERE name = 'Flaggenmast';

-- Länge und Abstand:
SELECT ST_Length(geo), ST_3DLength(geo),
       ST_Distance( ST_StartPoint(geo), ST_EndPoint(geo) ),
       ST_3DDistance( ST_StartPoint(geo), ST_EndPoint(geo) )
FROM GeoDbLand3D
WHERE name = 'Steg';

-- Test auf Schnitt und Verschneidung:
SELECT ST_Intersects (g1.geo, g2.geo), ST_3DIntersects (g1.geo, g2.geo),
       ST_Intersection (g1.geo, g2.geo), ST_3DIntersection (g1.geo, g2.geo)
FROM GeoDbLand3D g1, GeoDbLand3D g2
WHERE g1.name = 'Flaggenmast' AND g2.name = 'Land';

SELECT ST_Boundary(geo) AS rand3d, ST_Boundary(ST_Force2D(geo)) AS rand2d
FROM GeoDbLand3D
WHERE name = 'Steg';

SELECT ST_Force3D(geo,10) FROM GeoDbLand WHERE name = 'Steg';

SELECT ST_IsPlanar(ST_GeometryN(geo,1)), ST_IsPlanar(ST_GeometryN(geo,2))
FROM GeoDbLand WHERE name = 'Land'

INSERT INTO GeoDbLand3D VALUES (51, 'Hausdach',
ST_PolygonFromText('POLYGON Z ((9 9 6.5, 9 9.5 7, 9 10 6.5, 12 10 6.5, 
  12 9.5 7, 12 9 6.5, 9 9 6.5),
  (10 9.15 6.65, 11 9.15 6.65, 11 9.35 6.85, 10 9.35 6.85, 10 9.15 6.65))'));

SELECT ST_GeometryType(geo), ST_IsPlanar(geo), ST_IsValid(geo)
FROM GeoDbLand3D WHERE name = 'Hausdach';

UPDATE GeoDbLand3D SET geo = ST_MPolyFromText(
  'MULTIPOLYGON Z (((9 9 6.5, 9 9.5 7, 12 9.5 7, 12 9 6.5, 9 9 6.5),
   (10 9.15 6.65, 11 9.15 6.65, 11 9.35 6.85, 10 9.35 6.85, 10 9.15 6.65)),
   ((9 10 6.5, 12 10 6.5, 12 9.5 7, 9 9.5 7, 9 10 6.5)))')
WHERE name = 'Hausdach';

-- 12.3.1.2 Körper

SELECT Box3D(ST_Collect(geo)), ST_3DExtent(geo),
       ST_ZMin(ST_3DExtent(geo)), ST_ZMax(ST_3DExtent(geo))
FROM GeoDbLand3D;

INSERT INTO GeoDbLand3D
VALUES (52, 'Hauskörper', CAST('BOX3D(9 9 2, 12 10 6.5)' AS BOX3D));

SELECT ST_IsClosed(geo), geo FROM GeoDbLand3D WHERE name = 'Hauskörper';

UPDATE GeoDbLand3D SET geo = ST_Extrude(geo,0,0,-3)
WHERE name = 'Flaggenmast';

UPDATE GeoDbLand3D SET geo = ST_Extrude(geo,0,0,-1)
WHERE name = 'Land';

SELECT name, geo FROM GeoDbLand3D
WHERE name = 'Flaggenmast' OR name = 'Land' ORDER BY name;

SELECT name, ST_Volume(geo) FROM GeoDbLand3D WHERE ST_Dimension(geo) = 3;

SELECT name, ST_3DIntersects(geo, ST_MakePoint(9.5,9.25,6.75))
FROM GeoDbLand3D
WHERE name LIKE 'Haus%';

-- SELECT ST_3DIntersection(geo,CAST('BOX3D(2 3 0.5, 10 11 1.5)' AS BOX3D))
-- FROM GeoDbLand3D
-- WHERE name = 'Land';
-- Die Verbindung zur Datenbank wird aufgrund des Fehlers abgebrochen.

-- 12.3.2 Dreiecksvermaschungen

-- Punktdatentabelle anlegen: 
CREATE TABLE DGMPunkte (
  id       INTEGER,                     -- ID
  höhe     INTEGER,                     -- Höhenangabe
  punkt    GEOMETRY(POINTZ,25832),      -- Lage in ETRS 89/UTM 32N
  CONSTRAINT pk_dgmpkt PRIMARY KEY(id)  -- Primärschlüssel
);

-- Hier DGMPostGIS.sql ausführen.

UPDATE DGMPunkte SET punkt = ST_MakePoint(ST_X(punkt),ST_Y(punkt),höhe);

SELECT ST_DelaunayTriangles(ST_Collect(punkt),0.001,2) FROM DGMPunkte;

-- Bestimmung der Ausdehnung: 
SELECT ST_Extent(punkt) FROM DGMPunkte;     -- liefert BOX2D als Ergebnis
SELECT ST_SetSRID(ST_Extent(punkt),25832) FROM DGMPunkte;

-- Beispielhafte Gittererzeugung mit 10000 Meter Seitenlänge: 
SELECT (g).i, (g).j, (g).geom FROM
  (SELECT (ST_SquareGrid(10000, 
     (SELECT ST_SetSRID(ST_Extent(punkt),25832) FROM DGMPunkte))) AS g) sq;

-- Extraktion der Dreiecke aus der Vermaschung über ST_Dump:
SELECT (ST_Dump( ST_DelaunayTriangles(ST_Collect(punkt),0.001,2) )).geom
FROM DGMPunkte;

-- Zuordnung der Dreiecken der Vermaschung zu den Gitterzellen: 
SELECT (g).i, (g).j, count(t) FROM
  (SELECT (ST_Dump(ST_DelaunayTriangles(ST_Collect(punkt),0.001,2))).geom
   AS t FROM DGMPunkte) sq1
INNER JOIN
  (SELECT (ST_SquareGrid(10000, (SELECT ST_SetSRID(ST_Extent(punkt),25832)
   FROM DGMPunkte))) AS g) sq2
ON ST_Intersects( ST_Translate((g).geom,-1.234,-2.345), ST_Centroid(t) )
GROUP BY g
ORDER BY (g).i, (g).j

-- Erzeugung der Tabelle DGM aus der Anfrage: 
CREATE TABLE DGM AS
SELECT 'Z' || (g).i || '_' || (g).j AS zelle, ST_Collect(t) AS tin FROM
  (SELECT (ST_Dump(ST_DelaunayTriangles(ST_Collect(punkt),0.001,2))).geom
   AS t FROM DGMPunkte) sq1
INNER JOIN
  (SELECT (ST_SquareGrid(10000, (SELECT ST_SetSRID(ST_Extent(punkt),25832)
   FROM DGMPunkte))) AS g) sq2
ON ST_Intersects( ST_Translate((g).geom,-1.234,-2.345), ST_Centroid(t) )
GROUP BY g;

-- Abfragen: 
SELECT zelle, tin FROM DGM ORDER BY zelle;

SELECT sum(ST_NumGeometries(tin)) FROM DGM;

SELECT zelle, ST_Union(tin) FROM DGM
GROUP BY zelle;

ALTER TABLE DGM ADD CONSTRAINT pk_dgm PRIMARY KEY(zelle);

SELECT zelle, ST_MakePolygon(ST_Boundary((ST_Dump(tin)).geom)) FROM DGM;

SELECT ST_Tesselate(geo) FROM GeoDbLand3D
WHERE name = 'Hauskörper' OR name = 'Land';

-- Berechnung der 2D- und 3D-Oberfläche: 
SELECT ST_Area(tin), ST_3DArea(tin) FROM dgm WHERE zelle = 'Z58_526';

-- Test auf Schnitt in 2D und 3D: 
SELECT ST_Intersects( n, tin ),
  ST_3DIntersects( ST_Force3D(n,700), tin ) AS M700,
  ST_3DIntersects( ST_Force3D(n,800), tin ) AS M800
FROM DGM, (SELECT ST_Transform(Geometry(geog),25832) AS n
           FROM Ortsnetze WHERE vorwahl = '08325') AS sq
WHERE zelle = 'Z58_526';

-- Berechnung des Schnitts zwischen DGM und einem Ortsnetz als Ebene in 3D: 
SELECT ST_3DIntersection( ST_Force3D(ST_Transform(Geometry(geog),25832),800),
                          tin )
FROM Ortsnetze, DGM
WHERE vorwahl = '08325' AND zelle = 'Z58_526';

-- Berechnung des Schnitts zwischen DGM und Ortsnetz als Körper: 
SELECT ST_3DIntersection(ST_Extrude
       (ST_Force3D(ST_Transform(Geometry(geog),25832),800),0,0,-800), tin )
FROM Ortsnetze, DGM
WHERE vorwahl = '08325' AND zelle = 'Z58_526';

-- 12.3.3 Räumliche Anfragen

-- 3D-Punktanfrage (Filterschritt):
SELECT name FROM GeoDbLand3D
WHERE geo ~~ ST_MakePoint(5,5,2);

-- 3D-Rechteckanfrage = Quaderanfrage (Filterschritt):
SELECT name FROM GeoDbLand3D
WHERE geo &&& CAST('BOX3D(4 7.5 2.4, 15 16 5)' AS BOX3D);

-- 3D-Punktanfrage (Verfeinerungsschritt):
SELECT name FROM GeoDbLand3D
WHERE ST_Contains(geo,ST_MakePoint(5,5,2));   -- ST_3DIntersects nutzen!

-- 3D-Abstandsanfrage (Verfeinerungsschritt):
SELECT name FROM GeoDbLand3D
WHERE ST_3DDWithin(geo,ST_MakePoint(5,5,2),6);

-- Quaderanfrage (Verfeinerungsschritt):
SELECT name FROM GeoDbLand3D
WHERE ST_3DIntersects(geo, CAST('BOX3D(4 7.5 2.4, 15 16 5)' AS BOX3D));

CREATE INDEX DGM_tin_ix ON DGM USING GIST (tin gist_geometry_ops_nd);

SELECT * FROM DGM
WHERE tin &&& CAST('BOX3D(580000 5260000 800,590000 5270000 1200))' AS BOX3D);

EXPLAIN ANALYZE SELECT * FROM DGM
WHERE tin &&& CAST('BOX3D(580000 5260000 800,590000 5270000 1200))' AS BOX3D);

SELECT * FROM DGM
WHERE ST_3DIntersects(tin, ST_SetSRID(
   CAST('BOX3D(580000 5260000 800,590000 5270000 1200))' AS BOX3D),25832) );

CREATE TABLE DGMDreiecke AS
SELECT (ST_Dump(ST_DelaunayTriangles(ST_Collect(punkt),0.001,0))).geom AS d
FROM DGMPunkte ORDER BY d;

CREATE INDEX DGMDreiecke_d_brin_ix ON DGMDreiecke
USING BRIN (d brin_geometry_inclusion_ops_3d) WITH (pages_per_range = 8);

SELECT d FROM DGMDreiecke
WHERE d &&& CAST('BOX3D(580000 5260000 800,590000 5270000 1200))' AS BOX3D);

-- BRIN löschen und SP-GiST-Index anlegen: 
DROP INDEX DGMDreiecke_d_brin_ix;
CREATE INDEX DDGMDreiecke_d_spgist_ix
ON DGMDreiecke USING SPGIST (d spgist_geometry_ops_3d);

-- Quaderanfrage mit SP-GiST-Index ausführen: 
SELECT d FROM DGMDreiecke
WHERE d &/& CAST('BOX3D(580000 5260000 800,590000 5270000 1200))' AS BOX3D);

SELECT d FROM DGMDreiecke
WHERE ST_3DIntersects(d, ST_SetSRID(
   CAST('BOX3D(580000 5260000 800,590000 5270000 1200))' AS BOX3D),25832) );

EXPLAIN ANALYZE
SELECT d FROM DGMDreiecke
WHERE ST_3DIntersects(d, ST_SetSRID(
   CAST('BOX3D(580000 5260000 800,590000 5270000 1200))' AS BOX3D),25832) );

