-- ------------------------
-- KAPITEL 13 (PostGIS 3.1)
-- ------------------------

-- 13.1.3 PostGIS

SELECT ST_CollectionExtract(g,1), ST_ConvexHull(g) FROM
  (SELECT unnest(ST_ClusterWithin(ST_Transform(geom,25832),10000)) AS g
   FROM GeoGemeinden WHERE hektar < 1000) sq;

SELECT code, ST_ClusterKMeans(ST_Transform(geom,25832), 15) OVER () AS clid
FROM GeoGemeinden WHERE hektar < 1000;

SELECT clid, count(*), ST_Collect(geom), ST_ConvexHull(ST_Collect(geom)) FROM
  (SELECT code, geom, 
          ST_ClusterKMeans(ST_Transform(geom,25832), 15) OVER () AS clid
   FROM GeoGemeinden WHERE hektar < 1000) sq
GROUP BY clid;

SELECT clid, count(*), ST_Collect(geom), ST_ConvexHull(ST_Collect(geom)) FROM
  (SELECT code, geom, 
          ST_ClusterDBSCAN(ST_Transform(geom,25832),20000,15) OVER() AS clid
   FROM GeoGemeinden WHERE hektar < 1000) sq
GROUP BY clid
HAVING clid IS NOT NULL;   -- falls man die Ausreißer nicht haben möchte


-- 13.3.5 Trajektorien in PostGIS

CREATE TABLE Bahnen (
  id    NUMERIC(5),                      -- ID
  bahn  GEOMETRY(LINESTRINGM,25832),     -- Bahn über Maßzahlen
  CONSTRAINT bahnen_pk PRIMARY KEY (id)  -- Primärschlüssel
);

INSERT INTO Bahnen VALUES (1, 
  (SELECT ST_AddMeasure( ST_LineMerge(
            ST_LocateBetween(geom_m,stat_von,stat_nach,-20)), 100, 200)
   FROM BWSAbschnitte WHERE route_id = '0101_0052.248000') );
INSERT INTO Bahnen VALUES (2, 
  (SELECT ST_AddMeasure( ST_LineMerge(
            ST_LocateBetween(geom_m,stat_von,stat_nach,-40)), 50, 250)
   FROM BWSAbschnitte WHERE route_id = '0101_0052.248000') );
INSERT INTO Bahnen VALUES (3, 
  (SELECT ST_AddMeasure( ST_LineMerge(
            ST_LocateBetween(geom_m,stat_von,stat_nach,-30)), 120, 230)
   FROM BWSAbschnitte WHERE route_id = '0101_0052.248000') );

SELECT id, ST_IsValidTrajectory(bahn) AS valid, bahn FROM Bahnen;

-- Bestimmung der Zeitpunkte und Entfernungen der nächsten Annährung: 
SELECT b2.id, ST_ClosestPointOfApproach(b1.bahn, b2.bahn) AS cpa,
              ST_DistanceCPA(b1.bahn, b2.bahn) AS dist
FROM Bahnen b1, Bahnen b2
WHERE b1.id = 1 AND b2.id <> 1;

-- Bestimmung der Punkte der nächsten Annährung: 
SELECT ST_LocateAlong(bahn1,cpa) AS p1, ST_LocateAlong(bahn2,cpa) AS p2 FROM
  (SELECT b2.id, ST_ClosestPointOfApproach(b1.bahn, b2.bahn) AS cpa, 
          b1.bahn AS bahn1, b2.bahn AS bahn2
   FROM Bahnen b1, Bahnen b2
   WHERE b1.id = 1 AND b2.id <> 1) sq;

SELECT b1.id, b2.id, ST_DistanceCPA(b1.bahn, b2.bahn) AS dist
FROM Bahnen b1 INNER JOIN Bahnen b2
ON b1.id < b2.id AND ST_CPAWithin(b1.bahn, b2.bahn, 50);

