-- -------------------------
-- KAPITEL 2 (PostgreSQL 13)
-- -------------------------

-- 2.2: Nutzung ohne AutoCommit

-- 2.2.1 Datenmanipulation

-- Anlegen der Tabelle:
CREATE TABLE Gemeinden (
  gkz   DECIMAL(8),                    -- Gemeindeschlüssel
  name  VARCHAR(30)  NOT NULL,         -- Gemeindename
  einw  DECIMAL(7)   DEFAULT 0,        -- Einwohnerzahl
  CONSTRAINT pk_gem PRIMARY KEY (gkz)  -- Primärschlüssel
);
COMMIT;

-- Ändern der Tabelle:
ALTER TABLE Gemeinden ADD CONSTRAINT ck_gem_einw CHECK (einw >= 0);
COMMIT;

-- Löschen der Tabelle:
-- DROP TABLE Gemeinden;

-- 2.2.2.1 DML und DCL

-- Einfügen von Datensätze:
INSERT INTO Gemeinden (gkz,name,einw) VALUES (3241001,'Hannover',538068);
INSERT INTO Gemeinden (gkz,name) VALUES (3403000,'Oldenburg');
INSERT INTO Gemeinden VALUES (3451002,'Bad Zwischenahn',28896);
INSERT INTO Gemeinden VALUES (3451007,'Westerstede',22952);
INSERT INTO Gemeinden VALUES (3461004,'Elsfleth',9114);
INSERT INTO Gemeinden VALUES (3404000,'Osnabrück',165251);
INSERT INTO Gemeinden VALUES (3405000,'Wilhelmshaven',76089);
COMMIT;

-- Ändern vorhandener Datensätze:
UPDATE Gemeinden SET einw = 169077 WHERE gkz = 3403000;
UPDATE Gemeinden SET einw = einw - 1143 WHERE Name = 'Hannover';
COMMIT;

-- Löschen von Datensätzen und Rücknahme des Löschens:
DELETE FROM Gemeinden WHERE name = 'Hannover';
ROLLBACK;

-- 2.1.2.2 Erzeugung von Schlüsselwerten

-- Anlegen der Tabelle Kreise:
CREATE TABLE Kreise (
  id         INTEGER  GENERATED ALWAYS AS IDENTITY (START WITH 1),
                                              -- generierter Schlüssel
  name       VARCHAR(50)  NOT NULL,           -- Name des Kreises
  qkm        DECIMAL(8,2),                    -- Fläche in qkm
  kreisstadt DECIMAL(8),                      -- GKZ der Kreisstadt
  CONSTRAINT pk_id PRIMARY KEY (id),          -- Primärschlüssel
  CONSTRAINT ck_krs_qkm CHECK (qkm > 0)       -- Integritätsregel
);

-- Einfügen von Tupeln in die Tabelle Kreise:
INSERT INTO Kreise (name,qkm,kreisstadt)
  VALUES ('Oldenburg Stadt',102.97,3403000);
INSERT INTO Kreise (name,qkm,kreisstadt)
  VALUES ('Ammerland',728.16,3451007);
INSERT INTO Kreise (name,qkm,kreisstadt)
  VALUES ('Osnabrück Stadt',119.80,3404000);
INSERT INTO Kreise (name,qkm,kreisstadt)
  VALUES ('Osnabrück Land',2121.56,3404000);
COMMIT;

-- 2.1.3 Beziehungen

-- Ändern der Tabelle:
ALTER TABLE Kreise ADD CONSTRAINT fk_krs_kst FOREIGN KEY (kreisstadt)
  REFERENCES Gemeinden(gkz)                         -- Fremdschlüssel
COMMIT;

-- Anlegen der Tabelle Hochschulen:
CREATE TABLE Hochschulen (
  num    INTEGER,                      -- Hochschulschlüssel
  name   VARCHAR(20)  NOT NULL,        -- Hochschulname
  stud   DECIMAL(6),                   -- Zahl der Studierenden
  CONSTRAINT pk_hs PRIMARY KEY (num)   -- Primärschlüssel
);

-- Füllen der Tabelle:
INSERT INTO Hochschulen VALUES (1,'Uni Oldenburg',14445);
INSERT INTO Hochschulen VALUES (2,'Jade Hochschule',6639);
INSERT INTO Hochschulen VALUES (3,'Uni Hannover',28141);
COMMIT;

-- Anlegen der Vernüpfungstabelle HatStandortIn:
CREATE TABLE HatStandortIn (
  hs        INTEGER,                                -- Hochschulschlüssel
  gemeinde  DECIMAL(8),                             -- GKZ des Standorts
  CONSTRAINT pk_stdort PRIMARY KEY (hs, gemeinde),  -- komb. Primärschlüssel
  CONSTRAINT fk_stdort_hs FOREIGN KEY (hs)
    REFERENCES Hochschulen(num) ON DELETE CASCADE,  -- 1. Fremdschlüssel
  CONSTRAINT fk_stdort_gem FOREIGN KEY (gemeinde)
    REFERENCES Gemeinden(gkz) ON DELETE CASCADE     -- 2. Fremdschlüssel
);

-- Einfügen der Datensätze in HatStandortIn:
INSERT INTO HatStandortIn VALUES (3,3241001);
INSERT INTO HatStandortIn VALUES (1,3403000);
INSERT INTO HatStandortIn VALUES (2,3403000);
INSERT INTO HatStandortIn VALUES (2,3461004);
INSERT INTO HatStandortIn VALUES (2,3405000);
COMMIT;

-- 2.1.4.1 Einfache Anfragen

-- keine Projektion = Ausgabe aller Attribute:
SELECT * FROM Gemeinden;

-- Projektion und Umbenennung:
SELECT name, einw AS einwohner FROM Gemeinden;

-- Auswahl der Großstädte:
SELECT * FROM Gemeinden WHERE einw >= 100000;

-- Gemeinden, deren Name mit W beginnt oder mit weniger als 50000 Einw.:
SELECT name, einw
FROM Gemeinden
WHERE name LIKE 'W%' OR einw < 50000;

-- aufsteigende Sortierung nach dem Namen:
SELECT gkz, name FROM Gemeinden ORDER BY name;

-- absteigende Sortierung nach der Einwohnerzahl:
SELECT gkz, name FROM Gemeinden ORDER BY einw DESC;

-- Duplikate entfernen:
SELECT DISTINCT hs FROM HatStandortIn;

-- 2.1.4.2 Verbindende Anfragen

-- Die Kreisnamen mit den Namen der zugehörigen Kreisstädte:
SELECT Kreise.name, Gemeinden.name
FROM Kreise, Gemeinden
WHERE kreisstadt = gkz;

-- Verbund nach SQL-92:
SELECT Kreise.name, Gemeinden.name
FROM Kreise INNER JOIN Gemeinden
ON kreisstadt = gkz;

-- Äußerer Verbund:
SELECT Gemeinden.name, Kreise.name
FROM Gemeinden LEFT JOIN Kreise
ON gkz = kreisstadt;

-- Kartesisches Produkt:
SELECT Kreise.name, Gemeinden.name FROM Kreise, Gemeinden;
-- Kartesisches Produkt gemäß SQL-92:
SELECT Kreise.name, Gemeinden.name FROM Kreise CROSS JOIN Gemeinden;

-- Unteranfrage:
SELECT k.name AS kreis, g.name AS gemeinde
FROM Kreise k INNER JOIN (SELECT * FROM Gemeinden WHERE einw >= 100000) g
ON kreisstadt = gkz;

-- Fehlerhafte Unterabfrage: 
UPDATE Gemeinden SET einw = (SELECT num FROM Hochschulen)
WHERE name = 'Bad Zwischenahn';

-- IN-Bedingung:
SELECT name
FROM Gemeinden
WHERE gkz IN (SELECT kreisstadt FROM Kreise);

-- 2.1.4.3 Gruppierung und Aggregation

-- Anzahl der Gemeinden:
SELECT count(*) FROM Gemeinden;

-- Durchschnittliche Einwohnerzahl der größeren Gemeinden:
SELECT avg(einw) FROM Gemeinden WHERE einw > 50000;

-- Gruppierung:
SELECT gemeinde, count(hs)  -- der direkte Zugriff auf hs ist nicht zulässig
FROM HatStandortIn
GROUP BY gemeinde;

-- Gruppierung mit HAVING-Klausel:
SELECT gemeinde, count(hs)
FROM HatStandortIn
GROUP BY gemeinde
HAVING count(hs) > 1;

-- Window-Funktion:
SELECT gkz, name, einw, 100*einw/max(einw) OVER () AS proz FROM Gemeinden;

-- Window-Funktion mit PARTITION BY:
SELECT gkz, name, einw, floor(log(10,einw)) as klasse,
       avg(einw) OVER (PARTITION BY floor(log(10,einw)))
FROM Gemeinden;

-- 2.1.4.4 Sichten

-- Anlegen der Sicht:
CREATE VIEW Grossstädte AS 
  SELECT * FROM Gemeinden WHERE einw >= 100000;
COMMIT;

-- Verwenden der Sicht:
SELECT name, einw FROM Grossstädte;

-- Index anlegen:
CREATE INDEX Gemeinden_name_ix ON Gemeinden(name);
COMMIT;

-- Index wieder löschen:
DROP INDEX Gemeinden_name_ix;
COMMIT;

-- 2.2: Ab hier Nutzung im AutoCommit-Modus

-- 2.2.1.3 Logische Datenbankstruktur

SELECT * FROM Gemeinden;

SELECT * FROM public.Gemeinden;

-- Wechsel in das Schema information_schema: 
SET SCHEMA 'information_schema';
SELECT * FROM Gemeinden;   -- Fehler: Relation »gemeinden« existiert nicht
SELECT * FROM tables;      -- funktioniert

-- 2.2.2 SQL-92

-- PostgreSQL-Version abfragen: 
SELECT version();

SELECT * FROM
  (SELECT * FROM gemeinden WHERE einw >= 10000) sq
WHERE einw < 500000;

SELECT name, einw FROM Gemeinden
ORDER BY einw DESC
LIMIT 5;

-- 2.2.2.1 Datentypen

-- Umwandlung von gkz in eine Zeichenkette und Bestimmung von deren Länge: 
SELECT Length( CAST(gkz AS VARCHAR) ) FROM Gemeinden;
SELECT Length( gkz::VARCHAR ) FROM Gemeinden;

-- Tabelle zur Speicherung von Rasterkarten in BLOBs anlegen: 
CREATE TABLE Rasterkarten (
  id         INTEGER,                 -- ID
  name       VARCHAR(80)  NOT NULL,   -- Kartenname
  karte      BYTEA,                   -- Rasterkarte als BLOB
  worldfile  VARCHAR(255),            -- Georeferenzierung
  CONSTRAINT pk_rk PRIMARY KEY (id)   -- Primärschlüssel
);

-- 2.2.2.2 Materialisierte Sichten

CREATE MATERIALIZED VIEW GrossstädteMV AS 
  SELECT * FROM Gemeinden WHERE einw >= 100000;
  
INSERT INTO Gemeinden (gkz, name, einw) VALUES (4011000, 'Bremen', 567559);

SELECT * FROM GrossstädteMV;           -- ohne Bremen

REFRESH MATERIALIZED VIEW GrossstädteMV;

SELECT * FROM GrossstädteMV;           -- nun mit Bremen

-- 2.2.2.3 Feldtyp

-- Erweiterung der Tabelle "Gemeinden" um ein Feld:
ALTER TABLE Gemeinden
ADD pois VARCHAR(20) ARRAY[10];   -- oder: ARRAY oder []

UPDATE Gemeinden
SET pois = ARRAY['Marktplatz', 'Schloss', 'Lange Straße']
WHERE name = 'Oldenburg';

-- Abfrage:
SELECT name, pois[3]
FROM Gemeinden
WHERE CARDINALITY(pois) >= 3;         -- Prüfung, ob Element vorhanden

SELECT unnest(pois) AS poi FROM Gemeinden WHERE pois IS NOT NULL;

SELECT kreisstadt, array_agg(name) FROM kreise
GROUP BY kreisstadt;

-- 2.2.2.4 Verbundtyp

CREATE TYPE PunktTyp AS (
  id  NUMERIC(10),        -- Identifikationsnummer
  x NUMERIC(20,10),       -- x-Koordinate
  y NUMERIC(20,10)        -- y-Koordinate
);

CREATE TYPE PolygonTyp AS (
  id  NUMERIC(10),        -- Identifikationsnummer
  pts PunktTyp[]          -- Feld mit Punkten
);

ALTER TABLE Gemeinden ADD geom PolygonTyp;

UPDATE Gemeinden
SET geom = ROW(1234, ARRAY[ ROW(1, 8.214, 53.087)::PunktTyp,
                            ROW(2, 8.305, 53.148)::PunktTyp,
                            ROW(3, 8.173, 53.184)::PunktTyp,
                            ROW(4, 8.214, 53.087)::PunktTyp ])
WHERE name = 'Oldenburg';

SELECT (geom).id, (geom).pts[1], (geom).pts[1].x
FROM Gemeinden WHERE geom IS NOT NULL;

-- 2.2.2.5 Geometrische Datentypen

-- Tabelle um POINT ergänzen: 
ALTER TABLE Gemeinden ADD center POINT;

UPDATE Gemeinden
SET center = '(8.212993, 53.140025)'  -- oder: POINT(8.212993, 53.140025)
WHERE name = 'Oldenburg';

-- Erzeugung von weiteren Geometrien: 
SELECT LINE(POINT(8.214, 53.087),POINT(8.305, 53.148)),
       '{0.6703296703296748,-1.0,47.580912087912054}'::LINE;
	   
SELECT BOX(POINT(8.173,53.087),POINT(8.305,53.184)),
       CAST('(8.173,53.087),(8.305,53.184)' AS BOX);
	   
SELECT CIRCLE(POINT(8.214, 53.087),0.1), '<(8.214,53.087),0.1>'::CIRCLE; 

SELECT '((8.21,53.08),(8.30,53.14),(8.17,53.18),(8.21,53.08))'::POLYGON;

SELECT center[0], center[1] FROM Gemeinden WHERE center IS NOT NULL;

SELECT (BOX(POINT(8.173,53.087),POINT(8.305,53.184)))[0],
        (('(8.173,53.087),(8.305,53.184)'::BOX)[0])[0];

-- 2.2.2.6 JSON

CREATE TABLE GeoDbLandJson (
   id   DECIMAL(4),
   doc  JSONB       NOT NULL,
   CONSTRAINT GeoDbLandJson_pk PRIMARY KEY(id)
);

INSERT INTO GeoDbLandJson VALUES (1,
'{"id": 1,
  "name": "Flaggenmast", 
  "height": 10.5,
  "geo": {"type": "Point", "coordinates": [10,11]}}');
INSERT INTO GeoDbLandJson VALUES (2,
'{"id": 2,
  "name": "Steg", 
  "geo": {"type": "LineString", "coordinates": [[4,4],[5,7],[8,8]]}}');
INSERT INTO GeoDbLandJson VALUES (3,
'{"id": 3,
  "name": "Baumreihe",
  "species": ["Eiche", "Buche", "Eiche"],
  "geo": {"type": "MultiPoint", "coordinates": [[10,19],[12,19],[14,19]]},
  "photo": {"name": "Baumreihe", "file": "img64774.jpg", "type": "jpg",
            "date": "2020-10-27"}}');
INSERT INTO GeoDbLandJson VALUES (4,
'{"id": 4,
  "name": "Wall",
  "species": ["Kirschlorbeer", "Thuja"],
  "geo": {"type": "LineString", "coordinates":
          [[1,15],[2.414,15.586],[3,17],[3.586,18.414],[5,19],[7,19]]}}');
INSERT INTO GeoDbLandJson VALUES (5,
'{"id": 5,
  "name": "Haus",
  "geo": {"type": "Polygon", "coordinates":
          [[[9,9],[12,9],[12,10],[9,10],[9,9]]]}}');
INSERT INTO GeoDbLandJson VALUES (6,
'{"id": 6,
  "name": "Land",
  "geo": {"type": "MultiPolygon", "coordinates":
          [[[[1,1],[19,1],[19,20],[1,20],[1,1]],
            [[4,5],[3,10],[7,14],[13,15],[17,12],[16,8],[13,5],[9,3],[4,5]]],
           [[[7,7],[12,7],[13,11],[10,13],[7,10],[7,7]]]]}}');
		   
SELECT id, doc->'name' AS n, doc->'geo'->'type' AS gt FROM GeoDbLandJson;
SELECT id, doc->>'name' AS n, doc->'geo'->>'type' AS gt FROM GeoDbLandJson;

SELECT id, doc->'geo'->'coordinates'->0 AS e0 FROM GeoDbLandJson;
SELECT id, doc->'geo'->'coordinates'->0->0 AS e0 FROM GeoDbLandJson;

SELECT id, doc->'species'->0 AS e0 FROM GeoDbLandJson;
SELECT id, doc->'species'->>0 AS e0 FROM GeoDbLandJson;

SELECT id, doc#>'{geo,type}' AS geotype FROM GeoDbLandJson;

-- Rückgabe von JSONB: 
SELECT id, jsonb_extract_path(doc,'geo','type') AS type FROM GeoDbLandJson;
-- Rückgabe als Zeichenkette: 
SELECT id, jsonb_extract_path(doc,'geo','type')#>>'{}' AS type
FROM GeoDbLandJson;

-- Zulässige Vergleiche zwischen Zeichenketten: 
SELECT id FROM GeoDbLandJson WHERE doc->>'name' = 'Baumreihe';
SELECT id FROM GeoDbLandJson WHERE doc->'photo'->>'name'='Baumreihe';

-- Zulässige Vergleiche zwischen Zahlen: 
SELECT id FROM GeoDbLandJson WHERE CAST(doc->>'id' AS INTEGER) = 5;
SELECT id FROM GeoDbLandJson WHERE (doc->>'id')::INTEGER = 5;

SELECT id FROM GeoDbLandJson
WHERE doc->'geo'->'coordinates' @> CAST('[10,11]' AS JSONB);

SELECT id FROM GeoDbLandJson WHERE doc ? 'photo';
SELECT id FROM GeoDbLandJson WHERE doc ?| ARRAY['photo','height'];
SELECT id FROM GeoDbLandJson WHERE doc ?& ARRAY['photo','height'];

-- 2.2.2.7 Anfragepläne

EXPLAIN SELECT gemeinde, count(hs) FROM HatStandortIn GROUP BY gemeinde;
