-- -----------------------
-- KAPITEL 4 (PostGIS 3.1)
-- -----------------------

-- 4.1 PostGIS

CREATE EXTENSION postgis;           -- allgemeine Funktionalität
CREATE EXTENSION postgis_raster;    -- Unterstützung von Rasterdaten
CREATE EXTENSION postgis_topology;  -- Unterstützung topologischer Modelle
CREATE EXTENSION postgis_sfcgal;    -- 3D-Erweiterungen

SELECT version();           -- PostgreSQL-Version
SELECT PostGIS_Version();   -- PostGIS-Version

-- 4.1.1.1 Deklaration und Konstruktoren

CREATE TABLE GeoDbLand (
  id     INTEGER,                         -- ID
  name   VARCHAR(20),                     -- Objektname
  geo    GEOMETRY(GEOMETRY,0),            -- die Geometrie
  CONSTRAINT pk_gdbland PRIMARY KEY(id)   -- Primärschlüssel
);

-- Einfügen der Datensätze: 
INSERT INTO GeoDbLand (id,name,geo)
VALUES (1,'Flaggenmast',ST_PointFromText('POINT(10 11)'));

INSERT INTO GeoDbLand (id,name,geo)
VALUES (2,'Steg',ST_LineFromText('LINESTRING(4 4,5 7,8 8)', 0));

INSERT INTO GeoDbLand (id,name,geo)
VALUES (3,'Baumreihe',ST_MPointFromText('MULTIPOINT(10 19,12 19,14 19)'));

INSERT INTO GeoDbLand (id,name,geo)
VALUES (4,'Wall',ST_CurveToLine(
   ST_GeomFromText('COMPOUNDCURVE ( CIRCULARSTRING
     (1 15, 2.414 15.586, 3 17, 3.586 18.414, 5 19) , (5 19, 7 19) )') ));
	 
INSERT INTO GeoDbLand (id,name,geo)
VALUES (5,'Haus',ST_PolygonFromText('POLYGON((9 9,12 9,12 10,9 10,9 9))'));

INSERT INTO GeoDbLand (id,name,geo)
VALUES (6,'Land',ST_MPolyFromText(
  'MULTIPOLYGON( ( (1 1,19 1,19 20,1 20,1 1 ),
     (4 5,3 10,7 14,13 15,17 12,16 8,13 5,9 3,4 5) ),
   ( (7 7,12 7,13 11,10 13,7 10,7 7 ) ) )'));
   
SELECT f_table_name, f_geometry_column, coord_dimension, srid, type
FROM Geometry_Columns;

-- 4.1.1.2 Grundfunktionalität

-- Umwandlung in WKT (durch substring verkürzt) und WKB: 
SELECT name, substring(ST_AsText(geo) FROM 1 FOR 32), ST_AsBinary(geo)
FROM GeoDbLand;

-- Umwandlung in GML: 
SELECT name, ST_AsGML(geo) FROM GeoDbLand;

SELECT name, ST_AsGeoJSON(geo) FROM GeoDbLand;

SELECT ST_AsGeoJSON(t.*) FROM GeoDbLand t WHERE name = 'Flaggenmast';

SELECT name, GeometryType(geo) FROM GeoDbLand;
SELECT name, ST_GeometryType(geo) FROM GeoDbLand;

-- Rand: 
SELECT name, ST_Boundary(geo) FROM GeoDbLand;

-- Bestandteile eines Polygons: 
SELECT ST_X( ST_PointN( ST_InteriorRingN( ST_GeometryN(geo,1), 1), 3) )
FROM GeoDbLand WHERE name = 'Land';

-- Funktionen IsSimple und Dimension: 
SELECT name, ST_IsSimple(geo) FROM GeoDbLand
WHERE ST_Dimension(geo) >= 1;

SELECT name, ST_XMin(geo), ST_Points(geo) FROM GeoDbLand; 

-- Zugriff auf die Geometrie der Komponenten: 
SELECT name, (ST_Dump(geo)).geom FROM GeoDbLand;
-- Zugriff auf Indizes und Geometrie der Komponenten: 
SELECT name, (dump).path, (dump).geom FROM
  (SELECT name, ST_Dump(geo) AS dump FROM GeoDbLand) sq;
  
SELECT (dump).path, (dump).geom FROM
  (SELECT ST_DumpPoints(geo) AS dump FROM GeoDbLand WHERE name = 'Land') sq;
  
SELECT g, ST_CurveToLine(g), ST_CurveToLine(g,3),
       ST_CurveToLine(g,0.3,1), ST_CurveToLine(g,0.5,2) FROM 
  (SELECT ST_GeomFromText('CIRCULARSTRING(3 17,3.586 18.414,5 19)') AS g) sq;

SELECT g, ST_CurveToLine(g,0.3,1), ST_CurveToLine(ST_Reverse(g),0.3,1),
       ST_CurveToLine(g,0.3,1,1), ST_CurveToLine(ST_Reverse(g),0.3,1,1),
       ST_CurveToLine(g,0.3,1,3) FROM 
  (SELECT ST_GeomFromText('CIRCULARSTRING(3 17,3.586 18.414,5 19)') AS g) sq;

-- die bestehenden Geometrien überprüfen: 
SELECT ST_IsValid (geo) FROM GeoDbLand;

-- eine neue Geometrie überprüfen: 
SELECT (t).valid, (t).reason, (t).location FROM
  (SELECT ST_IsValidDetail(
          ST_GeomFromText('POLYGON((9 9,12 10,12 9,9 10,9 9))',0)) AS t) sq;

ALTER TABLE GeoDbLand
  ADD CONSTRAINT geodbl_geo_check CHECK (ST_IsValid(geo));
  
INSERT INTO GeoDbLand (id,name,geo) VALUES (7,'Haus',
  ST_GeomFromText('POLYGON((9 9,12 10,12 9,9 10,9 9))',0));
-- führt zu der Fehlermeldung: neue Zeile für Relation »geodbland« verletzt Check-Constraint »geodbl_geo_check«

-- 4.1.1.3 Geometrische Funktionen

-- Abstand: 
SELECT g1.name AS geo1, g2.name AS geo2, ST_Distance(g1.geo, g2.geo)
FROM GeoDbLand g1 INNER JOIN GeoDbLand g2
ON g1.id < g2.id
ORDER BY g1.id, g2.id;

-- minimal umgebendes Rechteck (MUR): 
SELECT geo, ST_Envelope(geo) FROM GeoDbLand WHERE name = 'Wall';

-- konvexe Hülle: 
SELECT geo, ST_ConvexHull(geo) FROM GeoDbLand WHERE name = 'Steg';

-- Differenz: 
SELECT ST_Difference(g1.geo, g2.geo)
FROM GeoDbLand g1, GeoDbLand g2
WHERE g1.name = 'Land' AND g2.name = 'Haus';

SELECT name, ST_Length(geo), ST_Area(geo) FROM GeoDbLand;

SELECT g2.name, ST_Distance(g1.geo,g2.geo),
       ST_DWithin(g1.geo,g2.geo,8), ST_DFullyWithin(g1.geo,g2.geo,8)
FROM GeoDbLand g1, GeoDbLand g2
WHERE g1.name = 'Flaggenmast' AND g2.name <> 'Flaggenmast';

-- Pufferzonen für das Haus: 
SELECT geo, ST_Buffer(geo,2) FROM GeoDbLand WHERE name='Haus';
SELECT geo, ST_Buffer(geo,2,3) FROM GeoDbLand WHERE name = 'Haus';
SELECT geo, ST_Buffer(geo,2,'join=bevel') FROM GeoDbLand WHERE name='Haus';

-- Pufferzonen für den Steg bzw. dessen konvexe Hülle: 
SELECT geo, ST_Buffer(ST_ConvexHull(geo), 2, 'join=mitre')
  FROM GeoDbLand WHERE name = 'Steg';
  
SELECT geo, ST_Buffer(ST_ConvexHull(geo), 2, 'join=mitre mitre_limit=1.5')
  FROM GeoDbLand WHERE name = 'Steg';
  
SELECT geo, ST_Buffer(geo, 2, 'quad_segs=6 endcap=flat')
  FROM GeoDbLand WHERE name = 'Steg';
  
SELECT geo, ST_Buffer(geo, 2, 'quad_segs=3 side=left')
  FROM GeoDbLand WHERE name = 'Steg';
  
-- 4.1.1.4 Topologische Beziehungen

SELECT ST_Intersects (g1.geo, g2.geo), ST_Within (g1.geo, g2.geo), 
       ST_Touches (g1.geo, g2.geo), ST_Contains (g1.geo, g2.geo),
       ST_Relate(g1.geo,g2.geo, 'T*F**F***')
FROM GeoDbLand g1, GeoDbLand g2
WHERE g1.name = 'Haus' AND g2.name = 'Land';

SELECT ST_Contains(geo,geo), ST_ContainsProperly(geo,geo)
FROM GeoDbLand WHERE name = 'Haus';

SELECT ST_Contains(geo,ST_Boundary(geo)), ST_Covers(geo,ST_Boundary(geo)),
       ST_Within(ST_Boundary(geo),geo), ST_CoveredBy(ST_Boundary(geo),geo)
FROM GeoDbLand WHERE name = 'Haus';

-- 4.1.1.5 Koordinatenbezugssysteme und -transformationen

SELECT auth_name, count(srid)
FROM Spatial_Ref_Sys
GROUP BY auth_name ORDER BY auth_name;

SELECT * FROM Spatial_Ref_Sys WHERE srid <> auth_srid;

SELECT srid, srtext FROM Spatial_Ref_Sys
WHERE auth_name = 'spatialreferencing.org';

-- Ausgabe ausgewählter EPSG-Bezugssysteme: 
SELECT srid, srtext FROM Spatial_Ref_Sys
WHERE srid IN (2056, 3416, 3857, 4326, 25832, 31467);

-- Gruppierung der Bezugssysteme nach Kategorie: 
SELECT substr(srtext,1,8), count(*)
FROM Spatial_Ref_Sys
WHERE srtext <> ''
GROUP BY substr(srtext,1,8)
ORDER BY substr(srtext,1,8);

-- Anlegen der Tabelle:
CREATE TABLE GeoGemeinden (
  code      VARCHAR(20),                 -- Gemeindeschlüssel
  name      VARCHAR(200)  NOT NULL,      -- Gemeindename
  status    VARCHAR(30),                 -- Status der Gemeinde
  wdcode    VARCHAR(15),                 -- Schlüssel für WikiData
  hektar    DECIMAL(10,2),               -- Fläche in Hektar
  zensus11  DECIMAL(7),                  -- Einwohnerzahl Zensus 2011
  einw19    DECIMAL(7),                  -- Einwohnerzahl Ende 2019
  geom      GEOMETRY(POINT,4326),        -- Gemeindezentrum in WGS84
  CONSTRAINT pk_ggem PRIMARY KEY (code)  -- Primärschlüssel
);

-- Hier GeoGemeindenPostGIS.sql ausführen.

SELECT DISTINCT ST_SRID(geom) FROM GeoGemeinden;

-- Pufferzone mit 0,5° um die Gemeindezentren in Nordrhein-Westfalen: 
SELECT ST_Buffer(geom,0.05) FROM GeoGemeinden WHERE code LIKE 'DE05%';

-- Abstand zwischen Berlin und Hamburg: 
SELECT ST_Distance(g1.geom,g2.geom)
FROM GeoGemeinden g1, GeoGemeinden g2
WHERE g1.name = 'Berlin' AND g2.name = 'Hamburg';

-- Transformation nach ETRS 89/UTM 32N: 
SELECT code, ST_AsText(geom) AS wgs,
             ST_AsText(ST_Transform (geom,25832)) AS etrs
FROM GeoGemeinden ORDER BY code;

-- Pufferzone mit 5000 m um die Gemeindezentren in Nordrhein-Westfalen: 
SELECT ST_Buffer(ST_Transform(geom,25832),5000)
FROM GeoGemeinden WHERE code LIKE 'DE05%';

-- Abstand zwischen Berlin und Hamburg in km: 
SELECT ST_Distance(ST_Transform(g1.geom,25832),
                   ST_Transform(g2.geom,25832))/1000 AS km
FROM GeoGemeinden g1, GeoGemeinden g2
WHERE g1.name = 'Berlin' AND g2.name = 'Hamburg';

-- 4.1.1.6 Geometrische Aggregatfunktionen

-- Aggregation der Gemeindezentren in Nordrhein-Westfalen: 
SELECT ST_Collect(geom) FROM GeoGemeinden
WHERE code LIKE 'DE05%';

-- Fläche des MUR und der konvexen Hülle der NRW-Gemeindezentren in km²: 
SELECT ST_Area(ST_Envelope(ST_Collect(ST_Transform(geom,25832))))/1000000,
       ST_Area(ST_ConvexHull(ST_Collect(ST_Transform(geom,25832))))/1000000
FROM GeoGemeinden
WHERE code LIKE 'DE05%';

-- Fläche der konvexen Hülle pro Landkreis: 
SELECT substring(code FROM 1 FOR 7) AS landkreis,
  ST_Area(ST_ConvexHull(ST_Collect(ST_Transform(geom,25832))))/1000000 AS qkm
FROM GeoGemeinden
WHERE code LIKE 'DE05%'
GROUP BY substring(code FROM 1 FOR 7)
HAVING count(geom) > 1
ORDER BY substring(code FROM 1 FOR 7);

SELECT ST_Union(geo) FROM GeoDbLand WHERE ST_Dimension(geo) = 2;

SELECT ST_XMin(box), ST_XMax(box), ST_YMin(box), ST_YMax(box) FROM
  (SELECT ST_Extent(geom) AS box FROM GeoGemeinden WHERE code LIKE 'DE05%') sq;

-- 4.1.2.1 Deklaration und Konstruktoren

CREATE TABLE Ortsnetze (
  vorwahl   VARCHAR(6),                     -- Vorwahlnummer
  name      VARCHAR(80),                    -- Name des Ortsnetzes
  geog      GEOGRAPHY(MULTIPOLYGON,4326),   -- Gebiet des Ortsnetzes
  CONSTRAINT pk_onetz PRIMARY KEY(vorwahl)  -- Primärschlüssel
);

-- Hier OrtsnetzePostGIS.sql ausführen.

SELECT * FROM Ortsnetze WHERE vorwahl LIKE '04%'

SELECT GEOMETRY(geog) FROM Ortsnetze;
SELECT GEOGRAPHY(geom) FROM GeoGemeinden;

SELECT f_table_name, f_geography_column, coord_dimension, srid, type
FROM Geography_Columns;

-- 4.1.2.2 Funktionalität

-- Abstand zwischen den Gebieten zweier Ortsnetze: 

SELECT ST_Distance(n1.geog, n2.geog)/1000 AS km
FROM Ortsnetze n1, Ortsnetze n2
WHERE n1.name = 'Berlin' AND n2.name = 'Hamburg';

-- Abstand zwischen den Zentren zweier Gemeinden: 
SELECT ST_Distance(GEOGRAPHY(g1.geom), GEOGRAPHY(g2.geom))/1000 AS km
FROM GeoGemeinden g1, GeoGemeinden g2
WHERE g1.name = 'Berlin' AND g2.name = 'Hamburg';

-- Fläche und Umfang in km² bzw. km: 
SELECT name, ST_Area(geog)/1000000 AS qkm, ST_Perimeter(geog)/1000 AS umfang
FROM Ortsnetze
WHERE name = 'Berlin' OR name = 'Hamburg';

-- Automatische Konvertierung von GEOMETRY in GEOGRAPHY: 
SELECT g.name, ST_Distance(n.geog, g.geom)/1000 AS dist_km
FROM Ortsnetze n, GeoGemeinden g
WHERE n.name = 'Frankfurt am Main' AND g.name LIKE 'Frankfurt%';

-- Einhalten eines Höchstabstands zum Ortsnetz Frankfurt am Main: 
SELECT g.name AS gemeinde, ST_DWithin(n.geog, g.geom, 250000) AS max_250km
FROM Ortsnetze n, GeoGemeinden g
WHERE n.name = 'Frankfurt am Main' AND g.einw19 >= 600000
ORDER BY g.name;

-- 4.1.3.1 Repräsentation

CREATE TABLE BWSAbschnitte (
  bws_id      VARCHAR(5),                      -- ID der Bundeswasserstraße
  stat_von    NUMERIC(10,6),                   -- Stationswert am Anfang
  stat_nach   NUMERIC(10,6),                   -- Stationswert am Ende
  route_id    VARCHAR(20),                     -- ID des Abschnitts
  stat_r      VARCHAR(10),                     -- Stationierungsrichtung
  geom        GEOMETRY(LINESTRING,25832),      -- Geometrie ohne Maßzahlen
  CONSTRAINT bwsabs_pk PRIMARY KEY (route_id)  -- Primärschlüssel
);

-- Hier BWSAbschnittePostGIS.sql ausführen.

SELECT * FROM BWSAbschnitte WHERE bws_id = '0409' ORDER BY route_id;

-- Tabelle um Geometrieattribut für Streckenzüge mit Maßzahlen ergänzen: 
ALTER TABLE BWSAbschnitte ADD geom_m GEOMETRY(LINESTRINGM,25832);

-- Geometrieattribut mithilfe von ST_AddMeasure belegen: 
UPDATE BWSAbschnitte SET geom_m = ST_AddMeasure(geom,stat_von,stat_nach)
                     WHERE stat_r = 'gleich';
UPDATE BWSAbschnitte SET geom_m = ST_AddMeasure(geom,stat_nach,stat_von)
                     WHERE stat_r = 'gegen';
					 
-- Bisheriges Geometrieattribut löschen: 
ALTER TABLE BWSAbschnitte DROP geom;

-- 4.1.3.2 Anfragen

SELECT route_id, stat_von, stat_nach, stat_r,
  ST_M(ST_StartPoint(geom_m)) AS m1, ST_ZMflag(ST_EndPoint(geom_m)) AS hasM
FROM BWSAbschnitte WHERE bws_id = '0409' ORDER BY route_id;

SELECT route_id, (d).path, ST_X((d).geom), ST_Y((d).geom), ST_M((d).geom) FROM
   (SELECT route_id, ST_DumpPoints(geom_m) AS d
    FROM BWSAbschnitte WHERE bws_id = '0409') sq
ORDER BY route_id, path;

-- Bestimmung eines Punktes über einen Stationswert: 
SELECT * FROM
  (SELECT route_id, ST_LocateAlong(geom_m,65) AS pos
   FROM BWSAbschnitte WHERE bws_id = '0101') sq
WHERE NOT ST_IsEmpty(pos);

-- Bestimmung eines Abschnitts zwischen zwei Stationswerten: 
SELECT * FROM
  (SELECT route_id, ST_LocateBetween(geom_m,66,78) AS seg
   FROM BWSAbschnitte WHERE bws_id = '0101') sq
WHERE NOT ST_IsEmpty(seg);

SELECT ST_InterpolatePoint (geom_m,
                            ST_PointFromText('POINT(534202 5847101)',25832))
FROM BWSAbschnitte WHERE route_id = '0101_0052.248000';

