-- -----------------------
-- KAPITEL 5 (PostGIS 3.1)
-- -----------------------

-- 5.5.1.1 Erzeugung eines räumlichen Index

-- Räumliche Indexe anlegen: 
CREATE INDEX GeoDbLand_geo_ix ON GeoDbLand USING GIST (geo);
CREATE INDEX GeoGemeinden_geom_ix ON GeoGemeinden USING GIST (geom);
CREATE INDEX Ortsnetze_geog_ix ON Ortsnetze USING GIST (geog);

SELECT tablename, indexname, indexdef
FROM PG_Indexes WHERE indexdef LIKE '% gist (%'

-- 5.5.1.2

-- Rechteckanfrage (Überlappung): 
SELECT name, geo, ST_GeomFromText('POLYGON((5 5,5 15,15 15,15 5,5 5))') AS r
FROM GeoDbLand
WHERE geo && ST_GeomFromText('POLYGON((5 5,5 15,15 15,15 5,5 5))');

-- Rechteckanfrage (Enthaltensein): 
SELECT name, geo
FROM GeoDbLand
WHERE geo @ ST_MakeBox2D( ST_MakePoint(5,5), ST_MakePoint(15,15) );
-- nachfolgend wird anstatt von ST_MakeBox2D ST_MakeEnvelope verwendet, also: ST_MakeEnvelope(5,5,15,15)

SELECT n1.name
FROM Ortsnetze n1 INNER JOIN Ortsnetze n2
ON n1.geog && n2.geog
WHERE n2.vorwahl = '04431';

EXPLAIN ANALYZE SELECT n1.name, n2.geog
FROM Ortsnetze n1 INNER JOIN Ortsnetze n2 ON n1.geog && n2.geog
WHERE n2.vorwahl = '04431';

SELECT Ortsnetze.name, GeoGemeinden.name
FROM Ortsnetze INNER JOIN GeoGemeinden
ON GEOMETRY(geog) ~ geom;

SELECT g1.name, ST_Distance(g1.geom,g2.geom)
FROM GeoGemeinden g1 INNER JOIN GeoGemeinden g2
ON g1.geom @ ST_Expand(g2.geom,0.175)
WHERE g2.code = 'DE03451002';

-- 5.5.1.3 Verfeinerungsschritt

-- Rechteckanfrage (Überlappung): 
SELECT name, geo, ST_GeomFromText('POLYGON((5 5,5 15,15 15,15 5,5 5))') AS r
FROM GeoDbLand
WHERE ST_Intersects(geo,
                    ST_GeomFromText('POLYGON((5 5,5 15,15 15,15 5,5 5))'));
					
-- Rechteckanfrage (Enthaltensein): 
SELECT name, geo
FROM GeoDbLand
WHERE ST_Within(geo,ST_MakeEnvelope(5,5, 15,15));

-- Rechteckanfrage (Enthaltensein mit ST_Relate): 
SELECT name, geo
FROM GeoDbLand
WHERE geo @ ST_MakeEnvelope(5,5, 15,15) AND
      ST_Relate(geo, ST_MakeEnvelope(5,5, 15,15), 'T*F**F***');

SELECT n1.name
FROM Ortsnetze n1 INNER JOIN Ortsnetze n2
ON ST_Intersects(n1.geog,n2.geog)
WHERE n2.vorwahl = '04431';

SELECT n1.name
FROM Ortsnetze n1 INNER JOIN Ortsnetze n2
ON n1.geog && n2.geog AND ST_Touches(Geometry(n1.geog),Geometry(n2.geog))
WHERE n2.vorwahl = '04431';

SELECT Ortsnetze.name, GeoGemeinden.name
FROM Ortsnetze INNER JOIN GeoGemeinden
ON ST_Contains(Geometry(geog),geom);

SELECT g1.name, ST_Distance(g1.geom,g2.geom)
FROM GeoGemeinden g1 INNER JOIN GeoGemeinden g2
ON ST_DWithin(g1.geom,g2.geom,0.175)
WHERE g2.code = 'DE03451002';
