-- -----------------------
-- KAPITEL 6 (PostGIS 3.1)
-- -----------------------

-- 6.3.4.1 Räumliche Ordnungen

-- Sortierung nach Hilbert-Ordnung: 
SELECT * FROM GeoGemeinden ORDER BY geom;

-- Erzeugung der Hilbert-Kurve: 
SELECT ST_MakeLine(geom), ST_Length(ST_MakeLine(geom)) FROM
  (SELECT geom FROM GeoGemeinden ORDER BY geom) sq
  
-- Gruppierung der Kreise nach GeoHash-Werten: 
SELECT ST_GeoHash(geometry(geog)) AS ghash, count(*) AS cnt, array_agg(name)
FROM Ortsnetze
GROUP BY ST_GeoHash(geometry(geog))
ORDER BY ST_GeoHash(geometry(geog));

-- 6.3.4.2 SP-GiST-Index

CREATE INDEX ortsnetze_geog_ix ON Ortsnetze USING SPGIST (geog);

SELECT * FROM Ortsnetze n1, Ortsnetze n2
WHERE n1.geog && n2.geog AND n2.vorwahl = '04431';

CREATE INDEX geogemeinden_geom_ix ON GeoGemeinden
USING SPGIST (POINT(ST_X(geom),ST_Y(geom)) kd_point_ops);

SELECT * FROM GeoGemeinden
WHERE POINT(ST_X(geom),ST_Y(geom)) <@ BOX(POINT(9,52),POINT(10,53));

-- 6.3.4.3 Block Range Index

-- Tabelle nach Hilbert-Ordnung sortiert erzeugen: 
CREATE TABLE GeoGemeinden4BRIN AS
  SELECT * FROM GeoGemeinden ORDER BY geom;
  
CREATE INDEX GeoGemeinden4BRIN_geom_ix ON GeoGemeinden4BRIN
USING BRIN (geom) WITH (pages_per_range = 8);

-- Auswahl aller Gemeinden in einem Anfragerechteck: 
SELECT name, geom
FROM GeoGemeinden4BRIN
WHERE geom @ ST_MakeEnvelope(8,53, 9,54);


-- 6.5.4.1 Indexerzeugung

-- Index von GeoGemeinden löschen und nebenläufig neu erzeugen: 
DROP INDEX GeoGemeinden_geom_ix;

CREATE INDEX CONCURRENTLY GeoGemeinden_geom_ix ON GeoGemeinden
USING GIST (geom);

-- 6.5.4.2 Index über GeoJSON

-- Nullwerte ausschließen: 
ALTER TABLE GeoGemeinden ALTER COLUMN geom SET NOT NULL;

-- Clusterindex erzeugen: 
CLUSTER GeoGemeinden_geom_ix ON GeoGemeinden;

-- GiST-Index erzeugen: 
CREATE INDEX GeoDbLandJson_geo_ix
ON GeoDbLandJson
USING GIST (ST_GeomFromGeoJSON(doc->'geo'));

-- Rechteckanfrage auf GeoJSON: 
SELECT id, ST_GeomFromGeoJSON(doc->'geo') AS geom
FROM GeoDbLandJson
WHERE ST_Intersects(ST_GeomFromGeoJSON(doc->'geo'),
        ST_SetSRID(ST_MakeBox2D( ST_MakePoint(5,5),
                                 ST_MakePoint(15,15) ), 4326));
-- bzw. mit ST_MakeEnvelope(5,5, 15,15, 4326)

EXPLAIN ANALYZE SELECT id, ST_GeomFromGeoJSON(doc->'geo') AS geom
FROM GeoDbLandJson
WHERE ST_Intersects(ST_GeomFromGeoJSON(doc->'geo'),
        ST_MakeEnvelope(5,5, 15,15, 4326));

